/* codescmp.c */

#include "stdinc.h"
#include <ctype.h>
#include <string.h>

#include "nls.h"
#include "strutil.h"
#include "chunks.h"
#include "asmdef.h"
#include "asmsub.h"
#include "asmpars.h"
#include "asmitree.h"  
#include "intpseudo.h"
#include "codevars.h"
#include "errmsg.h"

#include "codemn1610.h"

/*---------------------------------------------------------------------------*/

static CPUVar CPUMN1610;
static CPUVar CPUMN1613;

/*---------------------------------------------------------------------------*/

static Boolean Reg(const char *pAsc, Word *pErg)
{
	if (!strcasecmp(pAsc, "SP"))
	{
		*pErg = 5;
		return True;
	}

	if (!strcasecmp(pAsc, "STR"))
	{
		*pErg = 6;
		return True;
	}

	if (!strcasecmp(pAsc, "IC"))
	{
		*pErg = 7;
		return True;
	}

	if (strlen(pAsc) != 2)
	{
		return False;
	}

	if (pAsc[0] == 'R' && pAsc[1] >= '0' && pAsc[1] <= '4')
	{
		*pErg = pAsc[1] - '0';
		return True;
	}

	if (pAsc[0] == 'X' && pAsc[1] >= '0' && pAsc[1] <= '1')
	{
		*pErg = pAsc[1] - '0' + 3;
		return True;
	}

	return False;
}

static Boolean Skip(const char *pAsc, Word *pErg)
{
	static struct {
		char *str;
		Word skip;
	} SkipTable[] = {
		{ "SKP", 0x1 },
		{ "M",   0x2 },
		{ "PZ",  0x3 },
		{ "Z",   0x4 }, { "E",   0x4 },
		{ "NZ",  0x5 }, { "NE",  0x5 },
		{ "MZ",  0x6 },
		{ "P",   0x7 },
		{ "EZ",  0x8 },
		{ "ENZ", 0x9 },
		{ "OZ",  0xa },
		{ "ONZ", 0xb },
		{ "LMZ", 0xc },
		{ "LP",  0xd },
		{ "LPZ", 0xe },
		{ "LM",  0xf },
		{ NULL,  0x0 } };
	int i;

	for (i = 0; SkipTable[i].str; i++)
	{
		if (!strcasecmp(pAsc, SkipTable[i].str))
		{
			*pErg = SkipTable[i].skip;
			return True;
		}
	}

	return False;
}

static Boolean EM(const char *pAsc, Word *pErg)
{
	static struct {
		char *str;
		Word em;
	} EMTable[] = {
		{ "RE", 0x1 },
		{ "SE", 0x2 },
		{ "CE", 0x3 },
		{ NULL, 0x0 } };
	int i;

	for (i = 0; EMTable[i].str; i++)
	{
		if (!strcasecmp(pAsc, EMTable[i].str))
		{
			*pErg = EMTable[i].em;
			return True;
		}
	}

	return False;
}

static void DecodeAdr(Word Index)
{
	int idx;
	Word reg = 0;
	Word mode = 0;
	LongInt adr;
	Boolean OK;
	int disp = 0;
	int l;
	
	if ((Index & 0x0700) == 0x0600)	/* IMS/DMS */
	{
		if (!ChkArgCnt(1,1)) return;
		idx = 1;
	}
	else if ((Index & 0x0700) == 0x0700)	/* B/BAL */
	{
		if (!ChkArgCnt(1,1)) return;
		idx = 1;
	}
	else
	{
		if (!ChkArgCnt(2,2)) return;
		if (!Reg(ArgStr[1].Str, &reg)) return;
		if (reg == 6 || reg == 7) return;
		idx = 2;
	}

	l = strlen(ArgStr[idx].Str);

	fprintf (stderr, "EA = <%s>\n", ArgStr[idx].Str );

	if (l >= 4 && ArgStr[idx].Str[l - 1] == ')' && ArgStr[idx].Str[l-4] == '(')
	{
		tStrComp Left, Right;
		Word preg;

		StrCompSplitRef(&Left, &Right, &ArgStr[idx], ArgStr[idx].Str + l - 4);
		StrCompShorten(&Right, 1);
		fprintf (stderr, "Left = <%s> , Right = <%s>\n", Left.Str, Right.Str);
		if (!Reg(Right.Str, &preg)) return;
		fprintf (stderr, "preg = %d\n", preg);
		switch (preg)
		{
		case 3:
			mode = 0x4;
			break;
		case 4:
			mode = 0x5;
			break;
		case 7:
			mode = 0x1;
			break;
		default:
			return WrError(ErrNum_InvRegisterPointer);
		}
		disp = EvalStrIntExpression(&Left, Int8, &OK);
		fprintf (stderr, "disp = %d\n", disp);
		if (mode == 0x1)
		{
			if( disp > 0x7f || disp < -0x80) return WrError(ErrNum_OverRange);
		}
		else
		{
			if( disp > 0xff || disp < -0x00) return WrError(ErrNum_OverRange);
		}
	}
	else if (ArgStr[idx].Str[0] == '[')
	{
		if (l >= 2 && ArgStr[idx].Str[l - 1] == ']')
		{
			tStrComp Inner;

			StrCompAlloc(&Inner);
			StrCompCopySub(&Inner, &ArgStr[idx], 1, l - 2);
			fprintf (stderr, "Inner = <%s>\n", Inner.Str );

			l -= 2;
			if (l >= 4 && Inner.Str[l - 1] == ')' && Inner.Str[l - 4] == '(')
			{
				tStrComp Left, Right;
				Word preg;

				StrCompSplitRef(&Left, &Right, &Inner, Inner.Str + l - 4);
				StrCompShorten(&Right, 1);
				fprintf (stderr, "Left = <%s> , Right = <%s>\n", Left.Str, Right.Str);
				if (!Reg(Right.Str, &preg)) return;
				fprintf (stderr, "preg = %d\n", preg);
				if (preg != 7) return WrError(ErrNum_InvRegisterPointer);
				mode = 0x3;
				disp = EvalStrIntExpression(&Left, Int8, &OK);
				fprintf (stderr, "disp = %d\n", disp);
				if( disp > 0x7f || disp < -0x80) WrError(ErrNum_OverRange);
			}
			else 
			{
				adr = EvalStrIntExpression(&Inner, UInt16, &OK);
				fprintf (stderr, "adr = <%04X>\n", adr );
				if (!OK) return;

				if (adr < 0x100)
				{
					mode = 0x2;
					disp = adr;
				}
				else
				{
					adr -= EProgCounter();
					if( adr <= 0x7fL && adr >= -0x80L)
					{
						mode = 0x3;
						disp = adr & 0xff;
					}
					else WrError(ErrNum_JmpDistTooBig);
				}
			}
		}
	}
	else
	{
		adr = EvalStrIntExpression(&ArgStr[idx], UInt16, &OK);
		fprintf (stderr, "adr = <%04X>\n", adr );
		if (!OK) return;

		if (adr < 0x100)
		{
			mode = 0x0;
			disp = adr;
		}
		else
		{
			adr -= EProgCounter();
			if( adr <= 0x7fL && adr >= -0x80L)
			{
				mode = 0x1;
				disp = adr & 0xff;
			}
			else return WrError(ErrNum_JmpDistTooBig);
		}
	}

	fprintf (stderr, "mode = %d , reg = %d, disp = %02X\n", mode, reg, disp);
	
	WAsmCode[0] = Index | (mode << 11 ) | (reg << 8) | (disp & 0xff);
	CodeLen = 1;
}

static void DecodeFixed(Word Index)
{
	if (ChkArgCnt(0, 0))
	{
		WAsmCode[0] = Index;
		CodeLen = 1;
	}
}

static void DecodeRegImm8(Word Index)
{
	Word reg;
	Byte imm8;
	Boolean OK;
	
	if (ChkArgCnt(2, 2))
	{
		fprintf( stderr, "(%s) (%s)\n", ArgStr[1].Str, ArgStr[2].Str );

		if (Reg(ArgStr[1].Str, &reg))
		{
			imm8 = EvalStrIntExpression(&ArgStr[2], Int8, &OK);
			if (OK)
			{
				fprintf(stderr, "%02X %02X\n", reg, imm8 );

				WAsmCode[0] = Index | ((Word)reg << 8) | imm8;
				CodeLen = 1;
			}
		}
	}
}

static void DecodeReg(Word Index)
{
	Word reg;
	
	if (!ChkArgCnt(1,1)) return;

	if (!Reg(ArgStr[1].Str, &reg)) return;

	WAsmCode[0] = Index | (reg << 8);
	CodeLen = 1;
}

static void DecodeRegRegSkip(Word Index)
{
	Word reg1;
	Word reg2;
	Word skip = 0;

	if (!ChkArgCnt(2,3)) return;

	if (!Reg(ArgStr[1].Str, &reg1)) return;
	if (!Reg(ArgStr[2].Str, &reg2)) return;

	fprintf (stderr, "ArgCnt = %d\n", ArgCnt);
	
	if (ArgCnt == 3)
	{
		if (!Skip(ArgStr[3].Str, &skip)) return;
	}

	WAsmCode[0] = Index | (reg1 << 8) | reg2 | (skip << 4);
	CodeLen = 1;
}

static void DecodeShift(Word Index)
{
	Word reg;
	Word em;
	Word skip;

	if (!ChkArgCnt(1,3)) return;

	if (!Reg(ArgStr[1].Str, &reg)) return;

	fprintf (stderr, "ArgCnt = %d\n", ArgCnt);
	
	switch (ArgCnt)
	{
	case 1:
		em = 0;
		skip = 0;
		break;
	case 2:
		if (EM(ArgStr[2].Str, &em))
		{
			skip = 0;
		}
		else if (Skip(ArgStr[2].Str, &skip))
		{
			em = 0;
		}
		else return;
		break;
	case 3:
		if (!EM(ArgStr[2].Str, &em)) return;
		if (!Skip(ArgStr[3].Str, &skip)) return;
		break;
	default:
		return;
	}
	
	fprintf (stderr, "skip = %d , em = %d\n", skip, em);

	WAsmCode[0] = Index | (reg << 8) | (skip << 4) | em;
	CodeLen = 1;
}

static void DecodeBit(Word Index)
{
	Word reg;
	Word skip = 0;
	Byte imm4;
	Boolean OK;
	
	if (!ChkArgCnt(2,3)) return;

	if (!Reg(ArgStr[1].Str, &reg)) return;

	imm4 = EvalStrIntExpression(&ArgStr[2], Int8, &OK);
	if(!OK) return;

	if (ArgCnt == 3)
	{
		if (!Skip(ArgStr[3].Str, &skip)) return;
	}
	
	WAsmCode[0] = Index | (reg << 8) | (skip << 4) | imm4;
	CodeLen = 1;
}

static void DecodeLevel(Word Index)
{
	Word l;
	Boolean OK;
	
	if (!ChkArgCnt(1,1)) return;

	l = EvalStrIntExpression(&ArgStr[1], Int8, &OK);
	if (!OK) return;
	
	WAsmCode[0] = Index | l;
	CodeLen = 1;
}

/*---------------------------------------------------------------------------*/

static void AddAdr(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeAdr);
}
static void AddFixed(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeFixed);
}

static void AddRegImm8(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeRegImm8);
}

static void AddReg(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeReg);
}

static void AddRegRegSkip(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeRegRegSkip);
}

static void AddShift(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeShift);
}
	
static void AddBit(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeBit);
}
	
static void AddLevel(char *NName, Word NCode)
{
	AddInstTable(InstTable, NName, NCode, DecodeLevel);
}
	
static void InitFields(void)
{
	InstTable = CreateInstTable(48);

	AddAdr("L",   0xc000);
	AddAdr("ST",  0x8000);
	AddAdr("B",   0xc700);
	AddAdr("BAL", 0x8700);
	AddAdr("IMS", 0xc600);
	AddAdr("DMS", 0x8600);
	
	AddRegRegSkip("A",    0x5808);
	AddRegRegSkip("S",    0x5800);
	AddRegRegSkip("AND",  0x6808);
	AddRegRegSkip("OR",   0x6008);
	AddRegRegSkip("EOR",  0x6000);
	AddRegRegSkip("C",    0x5008);
	AddRegRegSkip("CB",   0x5000);
	AddRegRegSkip("MV",   0x7808);
	AddRegRegSkip("MVB",  0x7800);
	AddRegRegSkip("BSWP", 0x7008);
	AddRegRegSkip("DSWP", 0x7000);
	AddRegRegSkip("LAD",  0x6800);

	AddShift("SL", 0x200c);
	AddShift("SR", 0x2008);

	AddBit("SBIT", 0x3800);
	AddBit("RBIT", 0x3000);
	AddBit("TBIT", 0x2800);
	AddBit("AI",   0x4800);
	AddBit("SI",   0x4000);
	
	AddRegImm8("RD",  0x1800);
	AddRegImm8("WR",  0x1000);
	AddRegImm8("MVI", 0x0800);

	AddLevel("LPSW", 0x2004);
	AddFixed("H",   0x2000);
	AddReg("PUSH", 0x2001);
	AddReg("POP",  0x2002);
	AddFixed("RET", 0x2003);

}

static void DeinitFields(void)
{
	DestroyInstTable(InstTable);
}

/*---------------------------------------------------------------------------*/

static void MakeCode_MN1610(void)
{
	CodeLen = 0;
	DontPrint = False;

	if (Memo("")) return;

	if (DecodeIntelPseudo(False)) return;

	if (!LookupInstTable(InstTable, OpPart.Str))
		WrStrErrorPos(ErrNum_UnknownInstruction, &OpPart);
}

static Boolean IsDef_MN1610(void)
{
	return False;
}

static void SwitchFrom_MN1610(void)
{
	DeinitFields();
}

static void SwitchTo_MN1610(void)
{
	TurnWords = True;
	ConstMode = ConstModeC;

	PCSymbol = "*";
	HeaderID = 0x6e;
	NOPCode = 0x7808;	/* MV R0,R0 */
	DivideChars = ",";
	HasAttrs = False;

	ValidSegs = 1 << SegCode;
	Grans[SegCode] = 2;
	ListGrans[SegCode] = 2;
	SegInits[SegCode] = 0;
	SegLimits[SegCode] = 0xffff;

	MakeCode = MakeCode_MN1610;
	IsDef = IsDef_MN1610;
	SwitchFrom = SwitchFrom_MN1610;
	InitFields();
}

void codemn1610_init(void)
{
	CPUMN1610 = AddCPU("MN1610", SwitchTo_MN1610);
	CPUMN1613 = AddCPU("MN1613", SwitchTo_MN1610);
}
